from .file import CorsikaFile, CorsikaCherenkovFile, CorsikaParticleFile
from .version import __version__


__all__ = [
    'CorsikaFile',
    'CorsikaCherenkovFile',
    'CorsikaParticleFile',
    'as_dict',
    '__version__',
]


def as_dict(structured_array):
    '''
    Convert a structured array or row of a structure array to a python dict

    Useful for pretty printing run / event headers:
    >>> as_dict(event.header)
    '''
    return {k: structured_array[k] for k in structured_array.dtype.names}



from .dtypes import build_dtype, Field


run_end_fields = [
    Field(1, "RUNE", dtype="S4"),
    Field(2, "run_number", dtype="float32"),
    Field(3, "n_events", dtype="float32"),
]

run_end_dtype = build_dtype(run_end_fields)



from .dtypes import build_dtype, Field


event_end_fields = [
    Field(1, "event_end", dtype="S4"),
    Field(2, "event_number"),
    Field(3, "n_photons_weighted"),
    Field(4, "n_electrons_weighted"),
    Field(5, "n_hadrons_weighted"),
    Field(6, "n_muons_weighted"),
    Field(7, "n_particles_written"),
    Field(262, "chi_square_longitudinal"),
    Field(263, "n_photons_written"),
    Field(264, "n_electrons_written"),
    Field(265, "n_hadrons_written"),
    Field(266, "n_muons_written"),
    Field(267, "n_em_particles_preshower"),
]

event_end_dtype = build_dtype(event_end_fields)



from .dtypes import Field, build_dtype


longitudinal_header_fields = [
    Field(1, "LONG", shape="S4"),
    Field(2, "event_number", shape="f4"),
    Field(3, "particle_id", shape="f4"),
    Field(4, "total_energy", unit="GeV", shape="f4"),
    Field(5, "n_longitudinal", shape="f4"),
    Field(6, "longitudinal_id", shape="f4"),
    Field(7, "first_interaction_height", unit="g/cm2", shape="f4"),
    Field(8, "zenith", unit="rad", shape="f4"),
    Field(9, "azimuth", unit="rad", shape="f4"),
    Field(10, "energy_cutoff_hadrons", unit="GeV", shape="f4"),
    Field(11, "energy_cutoff_muons", unit="GeV", shape="f4"),
    Field(12, "energy_cutoff_electrons", unit="GeV", shape="f4"),
    Field(13, "energy_cutoff_photons", unit="GeV", shape="f4"),
]

longitudinal_data_fields = [
    Field(1, "vertical_depth", unit="g/cm2", shape="f4"),
    Field(2, "n_photons", shape="f4"),
    Field(3, "n_e_plus", shape="f4"),
    Field(4, "n_e_minus", shape="f4"),
    Field(5, "n_mu_plus", shape="f4"),
    Field(6, "n_mu_minus", shape="f4"),
    Field(7, "n_hadrons", shape="f4"),
    Field(8, "n_charged", shape="f4"),
    Field(9, "n_nuclei", shape="f4"),
    Field(10, "n_cherenkov", shape="f4"),
]

longitudinal_header_dtype = build_dtype(longitudinal_header_fields, itemsize=None)
longitudinal_data_dtype = build_dtype(longitudinal_data_fields, itemsize=None)



from collections import namedtuple, defaultdict
import numpy as np


Field = namedtuple("Field", ["position", "name", "unit", "shape", "dtype"])

# set defaults for unit, shape and dtype, this works for the 3 right most attributes
Field.__new__.__defaults__ = (None, 1, "f4")


def build_dtype(fields, itemsize=4 * 273):
    dt = defaultdict(list)
    if itemsize is not None:
        dt["itemsize"] = itemsize

    for field in fields:
        dt["names"].append(field.name)
        dt["offsets"].append((field.position - 1) * 4)
        if field.shape != 1:
            dt["formats"].append((field.dtype, field.shape))
        else:
            dt["formats"].append(field.dtype)

    return np.dtype(dict(**dt))




import numpy as np
import struct

from .run_header import run_header_types
from .run_end import run_end_dtype

from .event_header import event_header_types
from .event_end import event_end_dtype

from .data import cherenkov_photons_dtype, particle_data_dtype
from .longitudinal import longitudinal_data_dtype

from ..constants import RUNH_VERSION_POSITION, EVTH_VERSION_POSITION

__all__ = [
    "parse_event_header",
    "parse_run_header",
    "parse_cherenkov_photons",
    "parse_particle_data",
    "parse_longitudinal",
]


def parse_run_header(run_header_bytes):
    version = get_version(run_header_bytes, RUNH_VERSION_POSITION)
    # get corsika minor version (Truncate the float after first digit)
    version = float(str(version)[:3])
    return np.frombuffer(run_header_bytes, dtype=run_header_types[version])


def parse_run_end(run_end_bytes):
    return np.frombuffer(run_end_bytes, dtype=run_end_dtype)


def parse_event_header(event_header_bytes):
    version = get_version(event_header_bytes, EVTH_VERSION_POSITION)
    version = float(str(version)[:3])
    return np.frombuffer(event_header_bytes, dtype=event_header_types[version])


def parse_event_end(event_end_bytes):
    return np.frombuffer(event_end_bytes, dtype=event_end_dtype)


def get_version(header_bytes, version_pos):
    sl = slice(4 * (version_pos - 1), 4 * version_pos)
    return round(struct.unpack("f", header_bytes[sl])[0], 4)


def parse_data_block(data_block_bytes, dtype):
    data = np.frombuffer(data_block_bytes, dtype=dtype)
    empty = data == np.zeros(1, dtype=dtype)
    return data[~empty]


def parse_cherenkov_photons(data_block_bytes):
    return parse_data_block(data_block_bytes, dtype=cherenkov_photons_dtype)


def parse_particle_data(data_block_bytes):
    return parse_data_block(data_block_bytes, dtype=particle_data_dtype)


def parse_longitudinal(longitudinal_data_bytes):
    return parse_data_block(longitudinal_data_bytes, longitudinal_data_dtype)


def get_units_from_fields(subblock_fields):
    """Retrieve units as a dictionary from the fields of a file subblock.

    Dimensionless fields are not selected.

    Parameters
    ----------
    subblock_fields: list(Field)
        One of the defined lists of fields.

    Return
    ------
    units: dict
        Dictionary with field names as keys and
        string representations of units as values.
    """
    units = {field.name: field.unit for field in subblock_fields if field.unit is not None}
    return units



import gzip
import struct

from .constants import BLOCK_SIZE_BYTES


#: buffersize for files not written as fortran sequential file
DEFAULT_BUFFER_SIZE = BLOCK_SIZE_BYTES * 100
#: struct definition of the fortran record marker
RECORD_MARKER = struct.Struct('i')


def is_gzip(path):
    '''Test if a file is gzipped by reading its first two bytes and compare
    to the gzip marker bytes.
    '''
    with open(path, 'rb') as f:
        marker_bytes = f.read(2)

    return marker_bytes[0] == 0x1f and marker_bytes[1] == 0x8b


def is_zstd(path):
    '''Test if a file is compressed using zstd using its magic marker bytes
    '''
    with open(path, 'rb') as f:
        marker_bytes = f.read(4)

    return marker_bytes == b'\x28\xb5\x2f\xfd'


def open_compressed(path):
    if is_gzip(path):
        return gzip.open(path)

    if is_zstd(path):
        from zstandard import ZstdDecompressor
        return ZstdDecompressor().stream_reader(open(path, 'rb'))

    return open(path, 'rb')


def read_buffer_size(path):
    '''
    Reads the first 4 bytes of a file and checks if
    it is the 'RUNH' designation None is returned,
    if not interpret it as unsigned integer, the
    size of the CORSIKA buffer in bytes
    '''
    with open_compressed(path) as f:
        data = f.read(RECORD_MARKER.size)

        if data == b'RUNH':
            return None

        buffer_size, = RECORD_MARKER.unpack(data)

    return buffer_size


def iter_blocks(f):
    is_fortran_file = True
    buffer_size = DEFAULT_BUFFER_SIZE


    data = f.read(4)
    f.seek(0)
    if data == b'RUNH':
        is_fortran_file = False

    while True:
        # for the fortran-chunked output, we need to read the record size
        if is_fortran_file:
            data = f.read(RECORD_MARKER.size)
            if len(data) < RECORD_MARKER.size:
                raise IOError("Read less bytes than expected, file seems to be truncated")

            buffer_size, = RECORD_MARKER.unpack(data)

        data = f.read(buffer_size)

        n_blocks = len(data) // BLOCK_SIZE_BYTES
        for block in range(n_blocks):
            start = block * BLOCK_SIZE_BYTES
            stop = start + BLOCK_SIZE_BYTES
            block = data[start:stop]
            if len(block) < BLOCK_SIZE_BYTES:
                raise IOError("Read less bytes than expected, file seems to be truncated")
            yield block

        # read trailing record marker
        if is_fortran_file:
            f.read(RECORD_MARKER.size)


def read_block(f, buffer_size=None):
    '''
    Reads a block of CORSIKA output, e.g. 273 4-byte floats.

    Under some conditions, CORSIKA writes output as FORTRAN
    raw file format. This means, there is a 4-byte unsigned
    integer (the buffer size)
    before and after each block, which has to be skipped.
    '''
    if buffer_size is not None:
        pos = f.tell()
        if pos == 0:
            f.seek(4)

        if (pos + 4) % (buffer_size + 8) == 0:
            f.read(8)

    block = f.read(BLOCK_SIZE_BYTES)
    if len(block) < BLOCK_SIZE_BYTES:
        raise IOError("Read less bytes than expected, file seems to be truncated")

    return block



from .dtypes import build_dtype, Field

cherenkov_photons_fields = [
    Field(1, "n_photons", dtype="float32"),
    Field(2, "x", unit="cm", dtype="float32"),
    Field(3, "y", unit="cm", dtype="float32"),
    Field(4, "u", dtype="float32"),
    Field(5, "v", dtype="float32"),
    Field(6, "t", unit="ns", dtype="float32"),
    Field(7, "production_height", unit="cm", dtype="float32"),
]

cherenkov_photons_dtype = build_dtype(cherenkov_photons_fields, itemsize=None)

mmcs_cherenkov_photons_fields = [
    Field(1, "n_photons", dtype="float32"),
    Field(2, "x", unit="cm", dtype="float32"),
    Field(3, "y", unit="cm", dtype="float32"),
    Field(4, "u", dtype="float32"),
    Field(5, "v", dtype="float32"),
    Field(6, "t", unit="ns", dtype="float32"),
    Field(7, "wavelength", unit="nm", dtype="float32"),
    Field(8, "mother_particle", dtype="uint16"),
    Field(9, "production_height", unit="cm", dtype="float32"),
]

mmcs_cherenkov_photons_dtype = build_dtype(mmcs_cherenkov_photons_fields, itemsize=None)

particle_data_fields = [
    Field(1, "particle_description", dtype="float32"),
    Field(2, "px", unit="GeV/c", dtype="float32"),
    Field(3, "py", unit="GeV/c", dtype="float32"),
    Field(4, "pz", unit="GeV/c", dtype="float32"),
    Field(5, "x", unit="cm", dtype="float32"),
    Field(6, "y", unit="cm", dtype="float32"),
    Field(7, "t", unit="ns", dtype="float32"),
]

particle_data_dtype = build_dtype(particle_data_fields, itemsize=None)



import numpy as np
from collections import namedtuple

from .subblocks import (
    parse_run_header,
    parse_event_header,
    parse_event_end,
    parse_cherenkov_photons,
    parse_particle_data,
    parse_longitudinal,
    parse_run_end,
    get_version,
)
from .subblocks.longitudinal import longitudinal_header_dtype
from .subblocks.data import mmcs_cherenkov_photons_dtype
from .io import iter_blocks, read_buffer_size, open_compressed

from .constants import BLOCK_SIZE_BYTES, EVTH_VERSION_POSITION

Event = namedtuple('Event', ['header', 'data', 'longitudinal', 'end'])
PhotonEvent = namedtuple('PhotonEvent', ['header', 'photons', 'longitudinal', 'end'])
ParticleEvent = namedtuple(
    'ParticleEvent',
    ['header', 'particles', 'longitudinal', 'end']
)

def _to_floatarray(block):
    return np.frombuffer(block, dtype=np.float32)


class CorsikaFile:
    """
    A file to iterate over events in a CORSIKA binary output file.
    """

    def __init__(self, path, parse_blocks=True):
        self.EventClass = Event

        self.parse_blocks = parse_blocks
        self._buffer_size = read_buffer_size(path)
        self._f = open_compressed(path)
        self._block_iter = iter_blocks(self._f)

        runh_bytes = next(self._block_iter)
        if not runh_bytes[:4] == b'RUNH':
            raise ValueError('File does not start with b"RUNH"')

        self.run_header = parse_run_header(runh_bytes)[0]
        self.version = round(float(self.run_header['version']), 4)
        self._run_end = None

    @property
    def run_end(self):
        if self._run_end is None:
            pos = self._f.tell()

            if self._buffer_size is None:
                self._f.seek(0, 2)
            else:
                self._f.seek(-4, 2)

            self._f.seek(-BLOCK_SIZE_BYTES, 1)
            block = self._f.read(BLOCK_SIZE_BYTES)
            while block[:4] != b'RUNE':
                self._f.seek(-2 * BLOCK_SIZE_BYTES, 1)
                block = self._f.read(BLOCK_SIZE_BYTES)

            self._run_end = parse_run_end(block)[0]
            self._f.seek(pos)

        return self._run_end

    def __next__(self):
        block = next(self._block_iter)

        if block[:4] == b'RUNE':
            self._run_end = parse_run_end(block)
            raise StopIteration()

        if block[:4] != b'EVTH':
            raise IOError('EVTH block expected but found {}'.format(block[:4]))

        if self.parse_blocks:
            event_header = parse_event_header(block)[0]
        else:
            event_header = _to_floatarray(block)

        data_bytes = bytearray()
        long_bytes = bytearray()

        block = next(self._block_iter)
        while block[:4] != b'EVTE':

            if block[:4] == b'LONG':
                long_bytes += block[longitudinal_header_dtype.itemsize:]
            else:
                data_bytes += block

            block = next(self._block_iter)

        if self.parse_blocks:
            event_end = parse_event_end(block)[0]
            data = self.parse_data_blocks(data_bytes)
            longitudinal = parse_longitudinal(long_bytes)
        else:
            event_end = _to_floatarray(block)
            data = _to_floatarray(data_bytes).reshape(-1, 7)
            longitudinal = _to_floatarray(long_bytes)

        return self.EventClass(event_header, data, longitudinal, event_end)

    @classmethod
    def parse_data_blocks(cls, data_bytes):
        array = np.frombuffer(data_bytes, dtype='float32').reshape(-1, 7)
        return array[np.any(array != 0, axis=1)]

    def __iter__(self):
        return self

    def read_headers(self):
        pos = self._f.tell()
        self._f.seek(0)

        block_iter = iter_blocks(self._f)
        block = next(block_iter)
        event_header_data = bytearray()
        end_found = True

        while block:
            if block[:4] == b'RUNE':
                self._run_end = parse_run_end(block)[0]
                break

            if (
                block[:4] == b'EVTH'
                and get_version(block, EVTH_VERSION_POSITION) == self.version
            ):
                if not end_found:
                    raise IOError('Expected EVTE block before next EVTH')

                event_header_data += block

                end_found = False
            elif block[:4] == b'EVTE':
                end_found = True

            block = next(block_iter)

        self._f.seek(pos)

        event_headers = parse_event_header(event_header_data)

        return self.run_header, event_headers, self._run_end

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self._f.close()


class CorsikaCherenkovFile(CorsikaFile):

    def __init__(self, path, mmcs=False, **kwargs):
        super().__init__(path, **kwargs)

        self.EventClass = PhotonEvent
        self.mmcs = mmcs

    def parse_data_blocks(self, data_bytes):
        photons = parse_cherenkov_photons(data_bytes)
        if not self.mmcs:
            return photons

        mmcs = np.empty(len(photons), dtype=mmcs_cherenkov_photons_dtype)

        for col in ('x', 'y', 'u', 'v', 't', 'production_height'):
            mmcs[col] = photons[col]

        mmcs['n_photons'] = 1.0
        mmcs['wavelength'] = photons['n_photons'] % 1000
        mmcs['mother_particle'] = photons['n_photons'] // 100000

        return mmcs


class CorsikaParticleFile(CorsikaFile):

    def __init__(self, path, **kwargs):
        super().__init__(path, **kwargs)
        self.EventClass = ParticleEvent

    def parse_data_blocks(self, data_bytes):
        return parse_particle_data(data_bytes)



from collections import defaultdict
import warnings
from .dtypes import Field, build_dtype


run_header_fields_65 = [
    Field(1, "run_header", dtype="S4"),
    Field(2, "run_number"),
    Field(3, "date"),
    Field(4, "version"),
    Field(5, "n_observation_levels"),
    Field(6, "observation_height", unit="cm", shape=10),
    Field(16, "energy_spectrum_slope"),
    Field(17, "energy_min", unit="GeV"),
    Field(18, "energy_max", unit="GeV"),
    Field(19, "egs4_flag"),
    Field(20, "nkg_flag"),
    Field(21, "energy_cutoff_hadrons", unit="GeV"),
    Field(22, "energy_cutoff_muons", unit="GeV"),
    Field(23, "energy_cutoff_electrons", unit="GeV"),
    Field(24, "energy_cutoff_photons", unit="GeV"),
    Field(25, "physical_constants_and_interaction_flags", shape=50),
    Field(94 + 1, "cka", shape=40),
    Field(134 + 1, "ceta", shape=5),
    Field(139 + 1, "cstrba", shape=11),
    Field(254 + 1, "aatm", shape=5),
    Field(259 + 1, "batm", shape=5),
    Field(264 + 1, "catm", shape=5),
    Field(270, "nflain"),
    Field(271, "nfdif"),
    Field(272, "nflpi0_100nflpif"),
    Field(273, "nflche_100nfragm"),
]


# the corsika (up to and including version 7.69) user guide has an error here
# confirmed via mail by T. Pierog. NSHOW is at position 93, not 94 as stated
# in the manual
run_header_fields_7x = [
    Field(1, "run_header", dtype="S4"),
    Field(2, "run_number"),
    Field(3, "date"),
    Field(4, "version"),
    Field(5, "n_observation_levels"),
    Field(5 + 1, "observation_height", unit="cm", shape=10),
    Field(16, "energy_spectrum_slope"),
    Field(17, "energy_min", unit="GeV"),
    Field(18, "energy_max", unit="GeV"),
    Field(19, "egs4_flag"),
    Field(20, "nkg_flag"),
    Field(21, "energy_cutoff_hadrons", unit="GeV"),
    Field(22, "energy_cutoff_muons", unit="GeV"),
    Field(23, "energy_cutoff_electrons", unit="GeV"),
    Field(24, "energy_cutoff_photons", unit="GeV"),
    Field(24 + 1, "physical_constants_and_interaction_flags", shape=50),
    Field(75, "inclined_observation_plane_x", unit="cm"),
    Field(76, "inclined_observation_plane_y", unit="cm"),
    Field(77, "inclined_observation_plane_z", unit="cm"),
    Field(78, "inclined_observation_plane_theta", unit="deg"),
    Field(79, "inclined_observation_plane_phi", unit="deg"),
    Field(93, "n_showers"),  # Says 94 in the userguide, but it's 93
    Field(94 + 1, "cka", shape=40),
    Field(134 + 1, "ceta", shape=5),
    Field(139 + 1, "cstrba", shape=11),
    Field(248, "x_scatter", unit="cm"),
    Field(249, "y_scatter", unit="cm"),
    Field(254 + 1, "aatm", shape=5),
    Field(259 + 1, "batm", shape=5),
    Field(264 + 1, "catm", shape=5),
    Field(270, "nflain"),
    Field(271, "nfdif"),
    Field(272, "nflpi0_100nflpif"),
    Field(273, "nflche_100nfragm"),
]


run_header_dtype_7x = build_dtype(run_header_fields_7x)
run_header_dtype_65 = build_dtype(run_header_fields_65)


def warn_dtype():
    warnings.warn("Version unknown, using run header definition dtype of version 7.x")
    return run_header_dtype_7x

def warn_fields():
    warnings.warn("Version unknown, using run header fields definition of version 7.x")
    return run_header_fields_7x

run_header_fields = defaultdict(warn_fields)
run_header_fields[6.5] = run_header_fields_65
run_header_fields[7.4] = run_header_fields_7x
run_header_fields[7.5] = run_header_fields_7x
run_header_fields[7.6] = run_header_fields_7x
run_header_fields[7.7] = run_header_fields_7x

run_header_types = defaultdict(warn_dtype)
run_header_types[6.5] = run_header_dtype_65
run_header_types[7.4] = run_header_dtype_7x
run_header_types[7.5] = run_header_dtype_7x
run_header_types[7.6] = run_header_dtype_7x
run_header_types[7.7] = run_header_dtype_7x



import warnings
from collections import defaultdict

from .dtypes import Field, build_dtype

event_header_fields_65 = [
    Field(1, "event_header", dtype="S4"),
    Field(2, "event_number"),
    Field(3, "particle_id"),
    Field(4, "total_energy", unit="GeV"),
    Field(5, "starting_altitude", unit="g/cm2"),
    Field(6, "first_target_id"),
    Field(7, "first_interaction_height", unit="cm"),
    Field(8, "momentum_x", unit="GeV/c"),
    Field(9, "momentum_y", unit="GeV/c"),
    Field(10, "momentum_minus_z", unit="GeV/c"),
    Field(11, "zenith", unit="rad"),
    Field(12, "azimuth", unit="rad"),
    Field(13, "n_random_sequences"),
    Field(13 + 1, "random_seeds", shape=(10, 3)),
    Field(44, "run_number"),
    Field(45, "date"),
    Field(46, "version"),
    Field(47, "n_observation_levels"),
    Field(47 + 1, "observation_height", unit="cm", shape=10),
    Field(58, "energy_spectrum_slope"),
    Field(59, "energy_min", unit="GeV"),
    Field(60, "energy_max", unit="GeV"),
    Field(61, "energy_cutoff_hadrons", unit="GeV"),
    Field(62, "energy_cutoff_muons", unit="GeV"),
    Field(63, "energy_cutoff_electrons", unit="GeV"),
    Field(64, "energy_cutoff_photons", unit="GeV"),
    Field(65, "nflain"),
    Field(66, "nfdif"),
    Field(67, "nflpi0"),
    Field(68, "nflpif"),
    Field(69, "nflche"),
    Field(70, "nfragm"),
    Field(71, "earth_magnetic_field_x", unit="uT"),
    Field(72, "earth_magnetic_field_z", unit="uT"),
    Field(73, "egs4_flag"),
    Field(74, "nkg_flag"),
    Field(75, "low_energy_hadron_model"),
    Field(76, "high_energy_hadron_model"),
    Field(77, "cerenkov_flag"),
    Field(78, "neutrino_flag"),
    Field(79, "curved_flag"),
    Field(80, "computer"),
    Field(81, "theta_min", unit="deg"),
    Field(82, "theta_max", unit="deg"),
    Field(83, "phi_min", unit="deg"),
    Field(84, "phi_max", unit="deg"),
    Field(85, "cherenkov_bunch_size"),
    Field(86, "n_cherenkov_detectors_x"),
    Field(87, "n_cherenkov_detectors_y"),
    Field(88, "cherenkov_detector_grid_spacing_x", unit="cm"),
    Field(89, "cherenkov_detector_grid_spacing_y", unit="cm"),
    Field(90, "cherenkov_detector_length_x", unit="cm"),
    Field(91, "cherenkov_detector_length_y", unit="cm"),
    Field(92, "cherenkov_output_flag"),
    Field(93, "angle_array_x_magnetic_north", unit="rad"),
    Field(94, "additional_muon_information_flag"),
    Field(95, "egs4_multpliple_scattering_step_length_factor"),
    Field(96, "cherenkov_wavelength_min", unit="nm"),
    Field(97, "cherenkov_wavelength_max", unit="nm"),
    Field(98, "n_reuse"),
    Field(98 + 1, "reuse_x", shape=20),
    Field(118 + 1, "reuse_y", shape=20),
    Field(139, "sybill_interaction_flag"),
    Field(140, "sybill_cross_section_flag"),
    Field(141, "qgsjet_interaction_flag"),
    Field(142, "qgsjet_cross_section_flag"),
    Field(143, "dpmjet_interaction_flag"),
    Field(144, "dpmjet_cross_section_flag"),
    Field(145, "venus_nexus_epos_cross_section_flag"),
    Field(146, "muon_multiple_scattering_flag"),
    Field(147, "nkg_radial_distribution_range", unit="cm"),
    Field(148, "energy_fraction_if_thinning_level_hadronic"),
    Field(149, "energy_fraction_if_thinning_level_em"),
    Field(150, "actual_weight_limit_thinning_hadronic"),
    Field(151, "actual_weight_limit_thinning_em"),
    Field(152, "max_radius_radial_thinning_cutting", unit="cm"),
    Field(153, "viewcone_inner_angle", unit="deg"),
    Field(154, "viewcone_outer_angle", unit="deg"),
    Field(155, "transition_energy_low_high_energy_model", unit="GeV"),
]


event_header_fields_73 = event_header_fields_65 + [
    Field(156, "skimming_incidence_flag"),
    Field(157, "horizontal_shower_exis_altitude", unit="cm"),
    Field(158, "starting_height", unit="cm"),
    Field(159, "explicit_charm_generation_flag"),
    Field(160, "electromagnetic_subshower_hadronic_origin_output_flag"),
    Field(161, "conex_min_vertical_depth", unit="g/cm2"),
    Field(162, "conex_high_energy_treshold_hadrons", unit="GeV"),
    Field(163, "conex_high_energy_treshold_muons", unit="GeV"),
    Field(164, "conex_high_energy_treshold_em", unit="GeV"),
    Field(165, "conex_low_energy_treshold_hadrons", unit="GeV"),
    Field(166, "conex_low_energy_treshold_muons", unit="GeV"),
    Field(167, "conex_low_energy_treshold_em", unit="GeV"),
    Field(168, "observaton_level_curvature_flag"),
    Field(169, "conex_weight_limit_thinning_hadronic"),
    Field(170, "conex_weight_limit_thinning_em"),
    Field(171, "conex_weight_limit_sampling_hadronic"),
    Field(172, "conex_weight_limit_sampling_muons"),
    Field(173, "conex_weight_limit_sampling_em"),
]

event_header_fields_74 = event_header_fields_73.copy()

event_header_fields_75 = event_header_fields_74 + [
    Field(174, "augerhit_stripes_half_width", unit="cm"),
    Field(175, "augerhit_detector_distance", unit="cm"),
    Field(176, "augerhit_reserved"),
    Field(177, "n_multithin"),
    Field(177 + 1, "multithin_energy_fraction_hadronic", shape=6),
    Field(183 + 1, "multithin_weight_limit_hadronic", shape=6),
    Field(189 + 1, "multithin_energy_fraction_em", shape=6),
    Field(195 + 1, "multithin_weight_limit_em", shape=6),
    Field(199 + 3, "multithin_random_seeds", shape=(6, 3)),
    Field(220, "icecube_energy_threshold", unit="GeV"),
    Field(221, "icecube_gzip_flag"),
    Field(222, "icecube_pipe_flag"),
]
event_header_fields_76 = event_header_fields_75.copy()
event_header_fields_77 = event_header_fields_76.copy()

event_header_dtype_65xxx = build_dtype(event_header_fields_65)
event_header_dtype_73xxx = build_dtype(event_header_fields_73)
event_header_dtype_74xxx = build_dtype(event_header_fields_74)
event_header_dtype_75xxx = build_dtype(event_header_fields_75)
event_header_dtype_76xxx = build_dtype(event_header_fields_76)
event_header_dtype_77xxx = build_dtype(event_header_fields_77)


def warn_dtype():
    warnings.warn("Version unknown, using event header dtype definition of version 7.7XXX")
    return event_header_dtype_77xxx

def warn_fields():
    warnings.warn("Version unknown, using event header fields definition of version 7.7XXX")
    return event_header_fields_77

event_header_fields = defaultdict(warn_fields)
event_header_fields[6.5] = event_header_fields_65
event_header_fields[7.3] = event_header_fields_73
event_header_fields[7.4] = event_header_fields_74
event_header_fields[7.5] = event_header_fields_75
event_header_fields[7.6] = event_header_fields_76
event_header_fields[7.7] = event_header_fields_77

event_header_types = defaultdict(warn_dtype)
event_header_types[6.5] = event_header_dtype_65xxx
event_header_types[7.3] = event_header_dtype_73xxx
event_header_types[7.4] = event_header_dtype_74xxx
event_header_types[7.5] = event_header_dtype_75xxx
event_header_types[7.6] = event_header_dtype_76xxx
event_header_types[7.7] = event_header_dtype_76xxx
